#include <Wire.h>
#include <SPI.h>

#define LED  13
#define CE   10
#define SCL  13
#define SDA  11
#define Blank 9
//I2C Clock A5
//I2C Data  A4
#define HV_EN 8
#define B0 A0
#define B1 A1
#define B2 A2

#define testbit(data,bitno) ((data>>bitno)&0x001)    //Byte manipulation macros
#define bit_set(var,bitno) ((var) |= 1 << (bitno))
#define bit_clr(var,bitno) ((var) &= ~(1 << (bitno)))

bool t=1;
uint8_t g=0;
int cnt=0;
char k=0;

//RTC I2C Device Address (0x68, read 0xD1, write 0xD0)
#define RTCaddr 0x68
//RTC Register Addresses
#define SECS 0x00 //Seconds
#define MINS 0x01 //Minutes
#define HRS  0x02 //Hours
#define WKDY 0x03 //Weekday
#define DATE 0x04 //Date
#define MNTH 0x05 //Month
#define YRS  0x06 //Year

#define CTRL 0x0E //Control
#define STS  0x0F //Status
#define TME  0x10 //Time Offset, 2's complement (7 bit, 1 signbit), +/- 0.12ppm frequency change per 1 LSB (3.65s/year), + slow timebase, - speed timebase
#define TMP1 0x11 //Temperature MSB
#define TMP2 0x12 //Temperature LSB
#define SRAM 0x14 //Start of SRAM (14h-FFh)

unsigned char readI2C(unsigned char devaddr, unsigned char reg){
  unsigned char temp=0;
  //Serial.println("HERE!");
  Wire.beginTransmission(devaddr); //Read
  Wire.write(reg); //Word Address
  Wire.endTransmission();
  Wire.requestFrom(devaddr,1);
  temp=Wire.read();
  delayMicroseconds(10);
  return temp;
}

void writeI2C(unsigned char devaddr, unsigned char reg, unsigned char data){
  Wire.beginTransmission(devaddr); //Write
  Wire.write(reg);       //Word Address
  Wire.write(data);       //Data
  Wire.endTransmission(); //Send bytes and stop transmitting
  delayMicroseconds(10);
}

char encodeTime(int num){
  return ((num/10)<<4)|(num%10);
}

int decodeTime(char num){
  return ((10*((num&0x70)>>4))+(num&0x0F));
}

void softserial(int d0,int d1,int d2,int d3,int d4,int d5){
  digitalWrite(CE,LOW);
  unsigned char temp=0;
  //Serial.println("");

  for(int j=0;j<6;j++){
    if(j==0){temp=d5;}
    else if(j==1){temp=d4;}
    else if(j==2){temp=d3;}
    else if(j==3){temp=d2;}
    else if(j==4){temp=d1;}
    else{temp=d0;}

    //Serial.println(temp,BIN);
    
    for(int i=0;i<8;i++){
        digitalWrite(SCL,LOW);
        //delay(1);
        if(testbit(temp,i)){digitalWrite(SDA,HIGH);}else{digitalWrite(SDA,LOW);}
        //Serial.print(testbit(temp,i));
        //delay(1);
        digitalWrite(SCL,HIGH);
    }
  }
  digitalWrite(CE,HIGH);
  digitalWrite(SCL,LOW);
  digitalWrite(SDA,LOW);

  //Serial.println("");
}

                      //abcdefghijklmnop
                      //              xx
const uint16_t font[]{0b0000000000000000, //space
                      0b0000000000000000, //!
                      0b0100010000000000, //"
                      0b0000000000000000, //#
                      0b1011011100000000, //$
                      0b0000000000000000, //%
                      0b0000000000000000, //&
                      0b0000000001000000, //'
                      0b1001110000000000, //(
                      0b1111000000000000, //)
                      0b0000001111111100, //*
                      0b0000001101001000, //+
                      0b0000000000000000, //,
                      0b0000001100000000, //-
                      0b0000000000000010, //.
                      0b0000000000100100, // /
                      0b1111110000100100, //0
                      0b0110000000000000, //1
                      0b1101101100000000, //2
                      0b1111001100000000, //3
                      0b0110011100000000, //4
                      0b1011011100000000, //5
                      0b1011111100000000, //6
                      0b1110000000000000, //7
                      0b1111111100000000, //8
                      0b1111011100000000, //9
                      0b0000000000000000, //:
                      0b0000000000000000, //;
                      0b0000000000110000, //<
                      0b0001001100000000, //=
                      0b0000000010000100, //>
                      0b1100000100001000, //?
                      0b0000000000000000, //@
                      0b1110111100000000, //A
                      0b1111000101001000, //B
                      0b1001110000000000, //C
                      0b1111000001001000, //D
                      0b1001111000000000, //E
                      0b1000111000000000, //F
                      0b1011110100000000, //G
                      0b0110111100000000, //H
                      0b1001000001001000, //I
                      0b0111100000000000, //J
                      0b0000111000110000, //K
                      0b0001110000000000, //L
                      0b0110110010100000, //M
                      0b0110110010010000, //N
                      0b1111110000000000, //O
                      0b1100111100000000, //P
                      0b1111110000010000, //Q
                      0b1100111100010000, //R
                      0b1011011100000000, //S
                      0b1000000001001000, //T
                      0b0111110000000000, //U
                      0b0000110000100100, //V
                      0b0110110000010100, //W
                      0b0000000010110100, //X
                      0b0000000010101000, //Y
                      0b1001000000100100, //Z
                      0b1001110000000000, //[
                      0b0000000010010000, //\
                      0b0,
                      0b1111000000000000, //]
                      0b0100000000100000, //^
                      0b0001000000000000, //_                    
                      };

                    //abcdefghijklmnop
                    //              xx
uint16_t digits[16]{0b0000000000000000};  //16 character buffer to write to VFD in interrupt

void writeSPI(int g){ //send 5 bytes (40 bits) to fill serial HV shift register 
  uint16_t temp=1<<(15-g);
  //SPI.transfer16(val16)
  SPI.beginTransaction(SPISettings(8000000,LSBFIRST,SPI_MODE0));
  digitalWrite(CE,LOW);
  SPI.transfer(0b00000000);
  SPI.transfer(digits[g]&0xFF);
  SPI.transfer((digits[g]>>8)&0xFF);
  SPI.transfer(temp&0xFF);
  SPI.transfer((temp>>8)&0xFF);
  digitalWrite(CE,HIGH);
  SPI.endTransaction();
}

void printChar(char c,int p){
  digits[p]=font[c-32];
}

void writeString(String s,int l){
  unsigned char len=s.length();
  unsigned char p;
  for(int i=0;i<len;i++){
    p=s.charAt(i);
    p=toUpperCase(p);
    if(l+i<16){printChar(p,l+i);}
  }
}

ISR(TIMER2_OVF_vect){ //2kHz execution on timer2 overflow interrupt, write from buffer to vfd serial shift register
  if(g<15){ //scan 16 grids for each character
    g++;
  }else{
    g=0;
  }
  writeSPI(g);  //transfer segment data from buffer to shift register
}

void timer2init(void){  //initialize Timer 2 to trigger 2kHz interrupt
  TCCR2A = _BV(WGM21) | _BV(WGM20); //OC2A+B pins are regular IO
  TCCR2B = _BV(CS01) | _BV(CS00); //prescale by 32
  OCR2B = 127;
  TIMSK2=0x01;  //Timer2 Overflow Interrupt Enable
  asm("sei"); //Enable GLobal Interrupts
}

void setup() {
  pinMode(LED,OUTPUT);
  pinMode(A4,INPUT_PULLUP);
  pinMode(A5,INPUT_PULLUP);

  pinMode(B0,INPUT_PULLUP);
  pinMode(B1,INPUT_PULLUP);
  pinMode(B2,INPUT_PULLUP);

  pinMode(CE,OUTPUT);
  digitalWrite(CE,HIGH);
  pinMode(SCL,OUTPUT);
  pinMode(SDA,OUTPUT);
  pinMode(Blank,OUTPUT);
  digitalWrite(Blank,LOW);  //Disables outputs when high

  pinMode(HV_EN,OUTPUT);
  digitalWrite(HV_EN,HIGH);  //Turn on boost converter

  Wire.begin();
  //Wire.setClock(100000);
  Wire.setClock(400000);

  Serial.begin(115200);

  timer2init();

  writeString("Hello World",3);
}

void loop() {
  if(!digitalRead(B0))t^=1; //toggle boost converter by pressing B0
  digitalWrite(HV_EN,t);

  if(Serial.available()){ //write characters to display sent over uart
    k=toUpperCase(Serial.read());
    if(k!='\n'){
      if(cnt==0){
        writeString("                ",0);
      }
      printChar(k,cnt);
      cnt++;
    }else{
      printChar(' ',cnt-1);
      cnt=0;
    }
  }

  //SPCR&=~(1<<SPE);  //have to disable SPI peripheral to use D13 (SCK), be careful still updates shift register

  delay(1);

  /*if(cnt<1000){ //scroll through font set
    cnt++;
  }else{
    cnt=0;
    for(int i=0;i<16;i++){
      printChar(i+'*'+k,i);
    }
    if(k<53){k++;}else{k=0;}
  }*/
}
